import 'dart:convert';
import 'dart:developer' as dev;

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

/// Best logging utility
/// [logImpl] writes the log using [dev.log]
/// [logError], [logKV], [log] all implicitly calls [logImpl] only if [kReleaseMode]
mixin Loggable {
  static final Set<String> _mutedTags = {};
  String? _tag;

  /// [Indent] of the log
  final Indent indent = Indent(4);

  /// Log tag which is the runtime type of who implements this mixin by default
  String get tag => _tag ??= tagOf(this);

  /// Log an [error] and it's [stack]
  void logError(Object error, [StackTrace? stack]) => kReleaseMode ? null : logImpl('Runtime Error ⤵⤵⤵', error, stack);

  /// Log [key] -> [value]
  void logKV(dynamic key, dynamic value) => log('$key -> $value');

  /// Log a [message] (can be a [Printable])
  /// Optionally provide [details] callback to log more details automatically indented respectively to message
  void log(dynamic message, [VoidCallback? details]) {
    if (kReleaseMode || _mutedTags.contains(tag)) return;
    if (message is Printable) return message.print(this);
    LineSplitter.split(message.toString()).forEach(logImpl);
    if (details != null)
      try {
        indent.increase();
        details();
      } finally {
        indent.decrease();
      }
  }

  /// Actual log implementation, override this method to handle log write
  @protected
  void logImpl(String message, [Object? error, StackTrace? stack]) =>
      dev.log('$indent$message', name: tag, error: error, stackTrace: stack);

  static void mute<T>() => muteTag(T.toString());

  static void muteTag(String tag) => _mutedTags.add(tag);

  static String tagOf(Object object) => (object is State ? object.widget.runtimeType : object.runtimeType).toString();
}

/// Utility class that helps to provide indent spacing
/// Just through [toString] you can get the indentation of current level
class Indent {
  int _level = 1;

  /// The number of spaces per indent level
  final int step;

  /// The current indent level can be modified by invoking [increase] and [decrease]
  int get level => _level;

  /// Create Indent instance with given [step]
  Indent(this.step);

  /// Get indent by [level]
  String operator [](int level) => ' ' * (step * level);

  /// Increase level
  void increase() => _level++;

  /// Decrease level
  void decrease() => _level--;

  @override
  String toString() => this[_level];
}

/// Implement this interface to provide specific logging for an object
/// [print] is called when you pass this instance as message argument of [Loggable.log] e.g. log(myPrintable)
abstract class Printable {
  const Printable();

  /// Describe this instance using [_] [Loggable]
  /// Waring: calling _.log(this) will result in a stack overflow
  @protected
  void print(Loggable _);
}

/// Implementation of [Loggable] that allows to print logs into a String by calling [toString]
class LoggableStringBuffer with Loggable {
  final StringBuffer _buffer = StringBuffer();

  @override
  final String tag;

  LoggableStringBuffer({String? tag}) : tag = tag == null ? '' : '$tag:';

  @override
  void logImpl(String message, [Object? error, StackTrace? stack]) {
    _buffer.writeln('$tag$indent$message');
    if (error != null) LineSplitter.split(error.toString()).forEach(logImpl);
    if (stack != null) LineSplitter.split(stack.toString()).forEach(logImpl);
  }

  @override
  String toString() => _buffer.toString();
}
